import requests
import threading
from concurrent.futures import ThreadPoolExecutor
import batchquotes

thread_local = threading.local()

def get_session():
    #creates a session object for a thread if not initialized else using existing one
    if not hasattr(thread_local, 'session'):
        thread_local.session = requests.Session()
    return thread_local.session

def get_stock_quote(symbol):
    session = get_session()
    user_agent = batchquotes.USER_AGENT

    try:
        r = session.get(f"https://finance.yahoo.com/quote/{symbol}/history",
        headers={
                "User-Agent": user_agent
            })
    except:
        raise RuntimeError('Connection Error')

    if r.status_code == 302:
        raise RuntimeError(f'Invalid Symbol: {symbol}')
    #process request object 
    try:
        return batchquotes.extract_quote(r.text, symbol=symbol)
    except:
        raise RuntimeError(f'Invalid Symbol: {symbol}')

def get_quotes_threading(symbols, max_workers=None):
    assert type(symbols) == list
    #requests is not thread safe, so need to create a requests.session for each thread separately
    #thread_local looks like a global object in this namespace, but is specific to each individual thread

    with ThreadPoolExecutor(max_workers=max_workers) as executor:
        #map returns results in same order as list 
        quotes = executor.map(get_stock_quote, symbols)
    return list(quotes)



import asyncio 
import aiohttp
import batchquotes

#asyncio
async def fetch(session, symbol):
    url = f"https://finance.yahoo.com/quote/{symbol}/history"
    async with session.get(url, ssl=False) as response:
        if response.status != 200:
            response.raise_for_status()
        r = await response.text()
        return batchquotes.extract_quote(r, symbol=symbol)

async def coro(symbols):
    assert type(symbols) == list
    user_agent = batchquotes.USER_AGENT

    headers = {'User-Agent': user_agent}
    #define async func to get response object 
    async with aiohttp.ClientSession(headers=headers) as session:
        tasks = [fetch(session, symbol) for symbol in symbols]
        quotes = await asyncio.gather(*tasks)
    
    return quotes

def get_quotes_asyncio(symbols):
    return asyncio.run(coro(symbols))




import pytest 
from yfQuotes import get_quotes_asyncio, get_quotes_threading
import time

@pytest.fixture
def symbols():
    return [
        'GME', 'LRC-USD',
        'PLTR', 'NVDA', 'ROO.L', 'MMED.NE', 'ETH-USD',
        'HBAR-USD', 'DOGE-USD', 'BTC-USD', 'CTXR', 'F', 'AMC',
        'BB', 'CTXR', 'MRNA', 'BNTX', 'AMC', 'HBAR-USD',
        'OTLY', '^VIX', 'SPY'
    ]

def test_multithread(symbols):
    try:
        start = time.time()
        quotes = get_quotes_threading(symbols)
        end = time.time()
        print(quotes[0])
        print(f'threading: {end - start}')
        return True 
    except Exception as e:
        print(e)
        return False 

def test_asyncio(symbols):
    try:
        start = time.time()
        quotes = get_quotes_asyncio(symbols)
        end = time.time()
        print(quotes[0])
        print(f'asyncio: {end - start}')
        return True 
    except Exception as e:
        print(e)
        return False 


"""
module to decode protobuf messages from yahoo!finance

references these files: 
https://finance.yahoo.com/__finStreamer-worker.js --> shows base64 encoding
https://finance.yahoo.com/__finStreamer-proto.js --> shows .proto structure

for the decoding process of the streaming data 
"""

import base64
from .yfinance_pb2 import PricingData

sample_msg = 'CgdCVEMtVVNEFRKGR0cYoN6fvbJfIgNVU0QqA0NDQzApOAFFc/7EQEiAwKnKmQJVsSZIR12iFEVHZdAgOUVqC0JpdGNvaW4gVVNEsAGAwKnKmQLYAQTgAYDAqcqZAugBgMCpypkC8gEDQlRD+gENQ29pbk1hcmtldENhcIECAAAAwLkEckGJAgAAfF84FmxC'

class Decoder:
    #fields taken from fields outlines in .proto file 
    __fields__ = (
    'id', 'price', 'time', 'currency', 'exchange', 'quoteType',
    'marketHours', 'changePercent', 'dayVolume', 'dayLow', 'change',
    'shortName', 'lastSize', 'priceHint', 'vol_24hr', 'volAllCurrencies',
    'fromcurrency', 'lastMarket', 'circulatingSupply', 'marketcap')

    def __init__(self, msg):
        self.msg = msg
        self.msg_bytes = None
        #decode msg 
        self.decode()
        #parse msg 
        self.proto_obj = PricingData()
        try:
            self.proto_obj.ParseFromString(self.msg_bytes)
        except Exception as e:
            print(f'Parsing Error: {e}')
        
        #convert pricing data into a dict for access
        self.set_props()

    def decode(self):
        self.msg_bytes = base64.b64decode(self.msg)

    def set_props(self):
        for field in Decoder.__fields__:
            setattr(self, field, getattr(self.proto_obj, field))

    def __str__(self):
        return str(self.data)
    
    def __repr__(self):
        return f'Decoder({self.msg})'

    


import pytest 
from yfQuotes import Streamer
import time

@pytest.fixture
def symbols1():
    return [
        'GME', 'LRC-USD',
        'PLTR', 'NVDA', 'ROO.L', 'MMED.NE', 'ETH-USD',
    ]


def test_streamer(symbols1):
    s = Streamer()
    s.connect()
    assert s.socket.keep_running == True 

    s.add(symbols1)
    assert s.subscribed == set(symbols1)

    s.remove(symbols1)
    assert s.subscribed == set()

    s.disconnect()

    assert s.socket.keep_running == False


"""
Creating a websocket to connect to yfinance to retrieve live streaming quotes
- this is probably possible b/c the native website uses elements called fin-streamer
to constantly update prices. So in theory we can connect with a custom socket and query custom
prices???
"""
import gc
from typing import Iterable, Callable
import logging
import websocket
import threading
import ssl
import json 
from datetime import datetime
import time
from .decoder import Decoder

websocket.enableTrace(False)

#next step - see how to change subscribe dynamically? - basically need to find 
#a way to send a new msg during connection - don't use connection as context manager 
#and manually handle closing of connection

#use asyncio create_task to run async functions in background 
class Streamer:
    _socket_uri = 'wss://streamer.finance.yahoo.com/'
    _sslopt = {"cert_reqs": ssl.CERT_NONE}

    def __init__(self, subscribe: Iterable=None, on_message: Callable=None, on_error: Callable=None, 
                on_close: Callable=None, on_open: Callable=None, header: dict=None,
                run_in_background: bool=True):
        """
        Constructor for Streamer object
        Args:
        subscribe = list of yfinance symbols to subscribe to after connecting 
        on_connect: callback method for after connection initiation
        on_error: callback method for when error occurs
        on_close: callback method for connection close
        header: a dictionary for additional headers to be included 
        run_in_background: boolean to determine if non-blocking thread is created for process
        """
        self._logger = None 
        self._init_logger()
        self.subscribed = set(subscribe) if subscribe is not None else set()
        self.on_open = on_open if on_open else self.default_on_open
        self.on_error = on_error if on_error else self.default_on_error
        self.on_close = on_close if on_close else self.default_on_close
        self.header = header
        self.on_message = on_message if on_message else self.default_on_message
        self.run_in_background = run_in_background
        self.running = False 
        self._thread = None
        #initiate socket
        self.init_socket()


    def init_socket(self):
        self.socket = websocket.WebSocketApp(Streamer._socket_uri, on_message=self.on_message,
                                            on_error=self.on_error, on_close=self.on_close,
                                            on_open=self.on_open)
        self._logger.debug('WebSocketApp Initiated')
    
    def connect(self, wait=2):
        #by default doesnt use ssl certification
        if self.socket.keep_running:
            return

        self.running = True
        if self.run_in_background:
            #use _thread_run_forever here, so reconnection is automatic
            self._thread = threading.Thread(target=self._thread_run_forever)
            self._thread.daemon = True
            self._thread.start()
        else:
            self.socket.run_forever(sslopt=Streamer._sslopt)

        self._logger.debug('Connection initiated.') 
        #takes a second before socket is ready, so put a small wait time here
        time.sleep(wait)
        #initiate subscriptions
        if len(self.subscribed) and self.socket.keep_running:
            self.add(self.subscribed)
            self._logger.debug(f'initiated subscriptions: {self.subscribed}')

    def _thread_run_forever(self, retry_in: int=5):
        #if connection is closed, don't think currently run_forever reconnects itself automatically,
        #so doing this in a while loop in the thread to reconnect if an error occurs 
        while self.running:
            try:
                self.socket.run_forever(sslopt=Streamer._sslopt)
                self._logger.debug('Connection initiated.') 

            except KeyboardInterrupt:
                self.running = False 
                self._logger.debug('Keyboard Interrupt - Exiting loop')
            except Exception as e:
                gc.collect()
                self._logger.exception(f'Socket has disconnected: {e}')
            self._logger.debug(f'Reconnecting websocket in {retry_in} sec')
            time.sleep(retry_in)

    def disconnect(self):
        self.running = False 
        self.socket.close()
        if self.run_in_background:
            self._thread.join()
        self._logger.debug('Connection Closed.')


    def add(self, symbols):
        if symbols is None or len(symbols) == 0:
            return 

        #check overlaps with current subscriptions   
        new = set(symbols) - self.subscribed       
        new = set([s for s in symbols if s not in self.subscribed])
        if len(new) == 0:
            return 

        subscribe = {'subscribe': list(new)}
        msg = json.dumps(subscribe)
        try:
            self.socket.send(msg)
        except:
            self._logger.exception(f'Failed to subscribe to {new}')
        else:
            self.subscribed = self.subscribed | new
            self._logger.debug(f'subscribed to {new}')


    def remove(self, symbols):
        if symbols is None or len(symbols) == 0:
            return 
        
        exists = set(symbols) & self.subscribed
        if len(exists) == 0:
            return 
        
        unsubscribe = {'unsubscribe': list(exists)}
        msg = json.dumps(unsubscribe)
        try:
            self.socket.send(msg)
        except:
            self._logger.exception(f'Failed to send {msg}')
        else:
            self.subscribed = self.subscribed - exists 
            self._logger.debug(f'unsubscribed: {exists}')


    def _init_logger(self, stream=True, level=logging.DEBUG):
        self._logger = logging.getLogger('Streamer')
        formatter = logging.Formatter('%(asctime)s ; %(levelname)s ; %(module)s ; %(funcName)s ; %(message)s')
        file_handler = logging.FileHandler('stream.log', 'w')
        file_handler.setFormatter(formatter)
        self._logger.addHandler(file_handler)
        self._logger.setLevel(level)
        if stream:
            stream_handler = logging.StreamHandler()
            stream_handler.setFormatter(formatter)
            self._logger.addHandler(stream_handler)

    @staticmethod
    def default_on_message(app, message):
        quote = Quote(message)
        now = datetime.now().strftime('%H:%M:%S')
        print(f'{now} - {quote.id}: {quote.price}')
        return 

    @staticmethod
    def default_on_error(app, error):
        print(error)
        return

    @staticmethod
    def default_on_close(app, close_status_code, close_msg):
        print(f'close_status_code: {close_status_code}')
        print(f'close message: {close_msg}')
        return

    @staticmethod
    def default_on_open(app):
        pass

if __name__ == '__main__':
    s = Streamer(subscribe=['BTC-USD', 'LRC-USD'], run_in_background=True)
    s.connect()
    s.add(['0388.HK'])
    s.remove(['LRC-USD'])




    





from setuptools import setup, find_packages

VERSION = '0.0.4'
DESCRIPTION = 'get quotes from yahoo!Finance'
LONG_DESCRIPTION = "package to scrape live/stream quotes from yahoofinance"

setup(
    name="yfQuotes",
    version=VERSION,
    author='Justin Law',
    author_email="justin.kklaw@gmail.com",
    description=DESCRIPTION,
    long_description=LONG_DESCRIPTION,
    packages=find_packages(),
    install_requires=[
    'beautifulsoup4==4.9.3',
    'aiohttp==3.8.1',
    'websocket-client==1.2.3'],
    classifiers=[
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent"
    ],
    python_requires=">=3.7"   
)


"""
batchQuotes:

This module is inspired by the stockquotes package and scrapes yahoo finance
for live stock/crypto quotes. Amendments to the original package include only 
acquiring the current price. Users can choose between using concurrent.futures'
multithreading or asyncio to quickly retrieve a batch of quotes

"""
from bs4 import BeautifulSoup as bs 
import requests
import os
from .multithread import get_quotes_threading 
from .async_io import get_quotes_asyncio
import re 

USER_AGENT = 'Mozilla/5.0 (Macintosh; Intel Mac OS X 10_14_6) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/95.0.4638.69 Safari/537.36'

def extract_quote(r_text, symbol=None):

    def clean(text):
        #remove non-alphanum from beginning
        blacklist = '()%+'
        cleaned = text.strip(blacklist)
        cleaned = cleaned.replace(',', '')
        return cleaned

    soup = bs(r_text, features='lxml')

    if symbol is None:
        #if you don't pass symbol argument it'll try and find one
        meta = soup.find('meta', attrs={'content': re.compile('.')})
        symbol = meta['content'].split(',')[0]
    #get current price 
    data_streams = soup.find_all('fin-streamer', attrs={'data-symbol': symbol})
    ignore = ['marketState']
    data = dict(map(str, [s['data-field'], s.string]) for s in data_streams if s['data-field'] not in ignore)
    data = {key: clean(val) for key, val in data.items()}
    
    #format out_dict with more standard names
    new_keys = ['current_price', 'increase_dollars', 'increase_percent', 
    'regular_market_time', 'post_market_time', 'current_price_PM',
    'increase_dollars_PM', 'increase_percent_PM']

    new_dict = {}

    for new, old in zip(new_keys, data):
        new_dict[new] = data[old]

    return new_dict


#optional function to update user_agents 
# def get_latest_user_agents(os='macos', file_name=USER_AGENTS_PATH):
#     assert os in ['macos', 'chrome-os', 'ios', 'windows', 'android']
#     session = requests.session()
#     r = session.get(f'https://www.whatismybrowser.com/guides/the-latest-user-agent/{os}?utm_source=whatismybrowsercom&utm_medium=internal&utm_campaign=latest-user-agent-index')
#     soup = bs(r.text, features='lxml')
#     spans = soup.find_all('span', {"class": "code"})
#     agent_strings = [span.string for span in spans]
#     if file_name is not None:
#         with open(file_name, 'w') as f:
#             for string in agent_strings:
#                 f.write(string + '\n')
#     else:
#         return agent_strings








